import collections
from pathlib import Path

import customtkinter as ctk
import numpy as np


def fpe(filepath):
    return Path(filepath).expanduser().resolve()


def flatten_dict(input_dict, parent_key=False, separator="_"):
    items = []
    for key, value in input_dict.items():
        new_key = str(parent_key) + separator + key if parent_key else key
        if isinstance(value, collections.abc.MutableMapping):
            items.extend(flatten_dict(value, new_key, separator).items())
        elif isinstance(value, list):
            for k, v in enumerate(value):
                items.extend(flatten_dict({str(k): v}, new_key).items())
        else:
            items.append((new_key.upper(), value))
    return dict(items)


# https://github.com/matplotlib/matplotlib/issues/4556
def update_errorbar(errobj, x, y, xerr=None, yerr=None):
    ln, caps, bars = errobj

    if len(bars) == 2:
        assert (
            xerr is not None and yerr is not None
        ), "Your errorbar object has 2 dimension of error bars defined. You must provide xerr and yerr."
        barsx, barsy = bars  # bars always exist (?)
        try:  # caps are optional
            errx_top, errx_bot, erry_top, erry_bot = caps
        except ValueError:  # in case there is no caps
            pass

    elif len(bars) == 1:
        assert (xerr is None and yerr is not None) or (
            xerr is not None and yerr is None
        ), "Your errorbar object has 1 dimension of error bars defined. You must provide xerr or yerr."

        if xerr is not None:
            (barsx,) = bars  # bars always exist (?)
            try:
                errx_top, errx_bot = caps
            except ValueError:  # in case there is no caps
                pass
        else:
            (barsy,) = bars  # bars always exist (?)
            try:
                erry_top, erry_bot = caps
            except ValueError:  # in case there is no caps
                pass

    ln.set_data(x, y)

    try:
        errx_top.set_xdata(x + xerr)
        errx_bot.set_xdata(x - xerr)
        errx_top.set_ydata(y)
        errx_bot.set_ydata(y)
    except NameError:
        pass
    try:
        barsx.set_segments(
            [np.array([[xt, y], [xb, y]]) for xt, xb, y in zip(x + xerr, x - xerr, y)]
        )
    except NameError:
        pass

    try:
        erry_top.set_xdata(x)
        erry_bot.set_xdata(x)
        erry_top.set_ydata(y + yerr)
        erry_bot.set_ydata(y - yerr)
    except NameError:
        pass
    try:
        barsy.set_segments(
            [np.array([[x, yt], [x, yb]]) for x, yt, yb in zip(x, y + yerr, y - yerr)]
        )
    except NameError:
        pass


def error_bar_visibility(errobj, visible=True):
    ln, caps, bars = errobj

    ln.set_visible(visible)

    if len(bars) == 2:
        barsx, barsy = bars  # bars always exist (?)
        barsx.set_visible(visible)
        barsy.set_visible(visible)
        try:  # caps are optional
            errx_top, errx_bot, erry_top, erry_bot = caps
            errx_top.set_visible(visible)
            errx_bot.set_visible(visible)
            erry_top.set_visible(visible)
            erry_bot.set_visible(visible)
        except ValueError:  # in case there is no caps
            pass

    elif len(bars) == 1:
        (bars1d,) = bars  # bars always exist (?)
        bars1d.set_visible(visible)
        try:
            err_top, err_bot = caps
            err_top.set_visible(visible)
            err_bot.set_visible(visible)
        except ValueError:  # in case there is no caps
            pass


# From https://stackoverflow.com/questions/4140437/
class ValidateFloatVar(ctk.StringVar):
    """StringVar subclass that only allows valid float values to be put in it."""

    def __init__(self, master=None, value=None, name=None):
        ctk.StringVar.__init__(self, master, value, name)
        self._old_value = self.get()
        self.trace("w", self._validate)

    def _validate(self, *_):
        new_value = self.get()
        try:
            new_value == "" or float(new_value)
            self._old_value = new_value
        except ValueError:
            ctk.StringVar.set(self, self._old_value)



from pathlib import Path

import astropy.io.fits as pf
import astropy.units as u
import customtkinter as ctk
import matplotlib as mpl
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
from astropy.convolution import Gaussian1DKernel, convolve
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.visualization import (
    ImageNormalize,
    LinearStretch,
    LogStretch,
    ManualInterval,
    MinMaxInterval,
    PercentileInterval,
    SqrtStretch,
    make_lupton_rgb,
)
from astropy.wcs import WCS
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
from photutils.aperture import (
    CircularAperture,
    SkyCircularAperture,
    aperture_photometry,
)
from tqdm import tqdm

from pygcg.utils import (
    ValidateFloatVar,
    VerticalNavigationToolbar2Tk,
    error_bar_visibility,
    update_errorbar,
)


class SpecFrame(ctk.CTkFrame):
    def __init__(self, master, gal_id, **kwargs):
        super().__init__(master, **kwargs)

        if gal_id == "":
            return
        self.gal_id = gal_id
        self.plotted_components = dict(emission={}, absorption={})
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=0)
        self.grid_columnconfigure(1, weight=1)
        self.grid_columnconfigure(2, weight=1)

        self.plot_options_frame = ctk.CTkFrame(self)
        self.plot_options_frame.grid(row=0, column=0, columnspan=2, sticky="ew")

        self.redshift_plot = RedshiftPlotFrame(self, self.gal_id)
        self.redshift_plot.grid(row=1, column=2, sticky="news")

        self.emission_checkbox = ctk.CTkCheckBox(
            self.plot_options_frame, text="Emission", command=self.change_lines
        )
        self.emission_checkbox.grid(row=0, column=0, padx=20, pady=(10, 10), sticky="w")
        self.absorption_checkbox = ctk.CTkCheckBox(
            self.plot_options_frame, text="Absorption", command=self.change_lines
        )
        self.absorption_checkbox.grid(
            row=0, column=1, padx=20, pady=(10, 10), sticky="w"
        )

        self.redshift_frame = ctk.CTkFrame(self)
        self.redshift_frame.grid(row=2, column=2, sticky="news")
        self.redshift_frame.columnconfigure([0, 1, 2, 3, 4, 5], weight=1)
        self.redshift_label = ctk.CTkLabel(
            self.redshift_frame, text="Estimated redshift:"
        )
        self.redshift_label.grid(
            row=0, column=0, columnspan=2, padx=(20, 10), pady=(10,), sticky="w"
        )
        self.redshift_spec_info = ctk.CTkLabel(
            self.redshift_frame,
            text="test",
        )
        self.redshift_spec_info.grid(
            row=0,
            column=2,
            columnspan=2,
            padx=(20, 10),
            pady=(10,),
            sticky="we",
        )
        self.redshift_phot_info = ctk.CTkLabel(
            self.redshift_frame,
            text="test2",
        )
        self.redshift_phot_info.grid(
            row=0,
            column=4,
            columnspan=2,
            padx=(20, 10),
            pady=(10,),
            sticky="we",
        )
        self.current_redshift = ValidateFloatVar(
            master=self,
            value=0,
        )
        self.redshift_entry = ctk.CTkEntry(
            self.redshift_frame,
            textvariable=self.current_redshift,
        )
        self.redshift_entry.grid(
            row=1,
            column=0,
            columnspan=3,
            padx=(20, 10),
            pady=(10, 0),
            sticky="we",
        )
        self.redshift_entry.bind(
            "<Return>",
            self.update_lines,
        )
        self.reset_redshift_button = ctk.CTkButton(
            self.redshift_frame,
            text="Reset to grizli redshift",
            command=self.reset_redshift,
        )
        self.reset_redshift_button.grid(
            row=1,
            column=3,
            columnspan=3,
            padx=(20, 10),
            pady=(10, 0),
            sticky="we",
        )
        self.redshift_slider = ctk.CTkSlider(
            self.redshift_frame,
            from_=0,
            to=8,
            command=self.update_lines,
            number_of_steps=800,
        )
        self.redshift_slider.grid(
            row=2,
            column=0,
            columnspan=6,
            padx=(20, 10),
            pady=10,
            sticky="we",
        )

        self.z_q_checkbox_un = ctk.CTkCheckBox(
            self.redshift_frame,
            text="Unreliable z",
            command=self.z_q_update_un,
        )
        self.z_q_checkbox_un.grid(
            row=3, column=0, columnspan=2, padx=(20, 10), pady=10, sticky="w"
        )

        self.z_q_checkbox_te = ctk.CTkCheckBox(
            self.redshift_frame,
            text="Tentative z",
            command=self.z_q_update_te,
        )
        self.z_q_checkbox_te.grid(
            row=3, column=2, columnspan=2, padx=(20, 10), pady=10, sticky="w"
        )

        self.bad_seg_checkbox = ctk.CTkCheckBox(
            self.redshift_frame,
            text="Bad segmentation map",
            command=self.bad_seg_update,
        )
        self.bad_seg_checkbox.grid(
            row=3,
            column=4,
            columnspan=2,
            padx=(20, 10),
            pady=10,
            sticky="w",
        )

        self.muse_checkbox = ctk.CTkCheckBox(
            self.plot_options_frame,
            text="MUSE spectrum",
            command=self.change_components,
        )
        self.muse_checkbox.grid(
            row=0,
            column=2,
            padx=20,
            pady=(10, 10),
            sticky="w",
        )
        self.grizli_checkbox = ctk.CTkCheckBox(
            self.plot_options_frame,
            text="NIRISS spectrum",
            command=self.change_components,
        )
        self.grizli_checkbox.select()
        self.grizli_checkbox.grid(row=0, column=3, padx=20, pady=(10, 10), sticky="w")
        self.grizli_temp_checkbox = ctk.CTkCheckBox(
            self.plot_options_frame,
            text="Grizli templates",
            command=self.change_components,
        )
        self.grizli_temp_checkbox.grid(
            row=0, column=4, padx=20, pady=(10, 10), sticky="w"
        )
        self.grizli_temp_checkbox.select()

        self.images_frame = ImagesFrame(
            self, gal_id=self.gal_id, bg_color=self.cget("bg_color")
        )
        self.images_frame.grid(row=2, column=0, columnspan=2, sticky="news")

    def z_q_update_un(self):
        self._root().current_gal_data[
            "unreliable_redshift"
        ] = self.z_q_checkbox_un.get()

    def z_q_update_te(self):
        self._root().current_gal_data["tentative_redshift"] = self.z_q_checkbox_te.get()

    def bad_seg_update(self):
        self._root().current_gal_data["bad_seg_map"] = self.bad_seg_checkbox.get()

    def check_axes_colours(self):
        self.fig.set_facecolor("none")
        self.fig.canvas.get_tk_widget().config(background=self._root().bg_colour_name)
        self.nav_toolbar.config(background=self._root().bg_colour_name)
        self.nav_toolbar.winfo_children()[-2].config(
            background=self._root().bg_colour_name
        )
        self.nav_toolbar.winfo_children()[-1].config(
            background=self._root().bg_colour_name
        )
        self.nav_toolbar.update()

    def update_plot(self):
        if not hasattr(self, "pyplot_canvas"):
            self.gal_id = self._root().current_gal_id.get()

            self.fig = Figure(constrained_layout=True)
            self.pyplot_canvas = FigureCanvasTkAgg(
                figure=self.fig,
                master=self,
            )
            self.fig.canvas.get_tk_widget().config(
                background=self._root().bg_colour_name
            )

            self.fig_axes = self.fig.add_subplot(111)

            self.fig.canvas.mpl_connect("motion_notify_event", self.hover)

            self.nav_toolbar = VerticalNavigationToolbar2Tk(
                self.fig.canvas,
                self,
                pack_toolbar=False,
            )

            self.fig_axes.set_xlabel(r"Wavelength (${\rm \AA}$)")
            self.fig_axes.set_ylabel("Flux")

            self.custom_annotation = self.fig_axes.annotate(
                "", xy=(0, 0), xytext=(0, 0), textcoords="offset points"
            )
            self.custom_annotation.set_visible(False)

            self._update_all()

            f = zoom_factory(self.fig_axes)

            self.pyplot_canvas.get_tk_widget().grid(row=1, column=1, sticky="news")
            self.nav_toolbar.grid(row=1, column=0, sticky="news")

        if self.gal_id != self._root().current_gal_id.get():
            self.gal_id = self._root().current_gal_id.get()
            self._update_all()
        else:
            self._update_data()

        self.pyplot_canvas.draw_idle()
        self.update()
        self.fig.set_layout_engine("none")

    def _update_data(self):
        pad = self._root().config.get("catalogue", {}).get("seg_id_length", 5)

        # Check if *row data exists (small file, default grizli reduction)
        try:
            _row_path = [
                *(
                    Path(self._root().config["files"]["extractions_dir"])
                    .expanduser()
                    .resolve()
                ).glob(f"**/*{self._root().seg_id:0>{pad}}.row.fits")
            ][0]
            with pf.open(_row_path) as hdul:
                grizli_redshift = Table(hdul[1].data)["redshift"].value[0]
        except:
            try:
                # Check if *full [GLASS] or *maps [PASSAGE] files exist
                _full_path = [
                    *(
                        Path(self._root().config["files"]["extractions_dir"])
                        .expanduser()
                        .resolve()
                    ).glob(f"**/*{self._root().seg_id:0>{pad}}.full.fits")
                ] + [
                    *(
                        Path(self._root().config["files"]["extractions_dir"])
                        .expanduser()
                        .resolve()
                    ).glob(f"**/*{self._root().seg_id:0>{pad}}.maps.fits")
                ]
                _full_path = _full_path[0]
                with pf.open(_full_path) as hdul:
                    grizli_redshift = hdul[1].header["Z_MAP"].value[0]
            except:
                grizli_redshift = 0.0

            self.grizli_redshift = self._root().current_gal_data.get(
                "grizli_redshift", grizli_redshift
            )
            self.estimated_redshift = self._root().current_gal_data.get(
                "estimated_redshift", grizli_redshift
            )
            self.unreliable_redshift = self._root().current_gal_data.get(
                "unreliable_redshift", False
            )
            self.tentative_redshift = self._root().current_gal_data.get(
                "tentative_redshift", False
            )
            self.bad_seg = self._root().current_gal_data.get("bad_seg_map", False)

            self._root().current_gal_data["grizli_redshift"] = self.grizli_redshift
            self._root().current_gal_data[
                "estimated_redshift"
            ] = self.estimated_redshift
            self._root().current_gal_data[
                "unreliable_redshift"
            ] = self.unreliable_redshift
            self._root().current_gal_data[
                "tentative_redshift"
            ] = self.tentative_redshift
            self._root().current_gal_data["bad_seg_map"] = self.bad_seg
            self.current_redshift.set(self.estimated_redshift)
            self.redshift_slider.set(self.estimated_redshift)
            (
                self.z_q_checkbox_un.select()
                if self.unreliable_redshift
                else self.z_q_checkbox_un.deselect()
            )
            (
                self.z_q_checkbox_te.select()
                if self.unreliable_redshift
                else self.z_q_checkbox_te.deselect()
            )
            (
                self.bad_seg_checkbox.select()
                if self.bad_seg
                else self.bad_seg_checkbox.deselect()
            )

        try:
            phot_z_name = self._root().config.get("catalogue", {}).get("zphot", "zphot")
            phot_z = self._root().tab_row[phot_z_name]
            assert phot_z != "--"
            phot_z_text = f"z\u209a\u2095\u2092\u209c = {phot_z:.3f} "
        except Exception as e:
            phot_z_text = "z\u209a\u2095\u2092\u209c = n/a  "
        self.redshift_phot_info.configure(
            text=(f"{phot_z_text}"),
        )

        try:
            spec_z_name = self._root().config.get("catalogue", {}).get("zspec", "zspec")
            spec_z = self._root().tab_row[spec_z_name]
            assert spec_z != "--"
            spec_z_text = f"z\u209b\u209a\u2091\U0001E03F = {spec_z:.3f} "
        except Exception as e:
            spec_z_text = f"z\u209b\u209a\u2091\U0001E03F = n/a  "
        self.redshift_spec_info.configure(
            text=(f"{spec_z_text}"),
        )

    def _update_all(self):
        self.check_axes_colours()
        self.fig.set_layout_engine("constrained")

        self._update_data()

        self.cube_path = (
            Path(self._root().config["files"]["cube_path"]).expanduser().resolve()
        )
        if not self.cube_path.is_file():
            self.muse_checkbox.configure(state="disabled")
        else:
            self.muse_checkbox.configure(state="normal")

        if self.grizli_checkbox.get():
            self.plot_grizli()
        if self.grizli_temp_checkbox.get():
            self.plot_grizli(templates=True)
        if self.muse_checkbox.get():
            self.plot_MUSE_spec()

        self.redshift_plot.update_z_grid()
        self.change_lines()
        self.images_frame.update_images()

    def plot_grizli(self, templates=False):
        pad = self._root().config.get("catalogue", {}).get("seg_id_length", 5)
        file_path = [
            *(
                Path(self._root().config["files"]["extractions_dir"])
                .expanduser()
                .resolve()
            ).glob(f"**/*{self._root().seg_id:0>{pad}}.1D.fits")
        ] + [
            *(
                Path(self._root().config["files"]["extractions_dir"])
                .expanduser()
                .resolve()
            ).glob(f"**/*{self._root().seg_id:0>{pad}}.spec1D.fits")
        ]
        file_path = file_path[0]

        if templates:
            dict_key = "grism_templates"
        else:
            dict_key = "grisms"

        ymax = 0
        colours = {
            self._root().filter_names[0]: "C2",
            self._root().filter_names[1]: "C1",
            self._root().filter_names[2]: "C0",
        }

        if dict_key not in self.plotted_components.keys():
            self.plotted_components[dict_key] = dict()
        with pf.open(file_path) as hdul:
            for hdu in hdul[1:]:
                data_table = Table(hdu.data)
                clip = data_table["err"] > 0
                if clip.sum() == 0:
                    clip = np.isfinite(data_table["err"])
                if templates:
                    try:
                        self.plotted_components[dict_key][hdu.name].set_data(
                            data_table["wave"][clip],
                            data_table["line"][clip] / data_table["flat"][clip] / 1e-19,
                        )
                    except:
                        (self.plotted_components[dict_key][hdu.name],) = (
                            self.fig_axes.plot(
                                data_table["wave"][clip],
                                data_table["line"][clip]
                                / data_table["flat"][clip]
                                / 1e-19,
                                c="red",
                                alpha=0.7,
                                zorder=10,
                            )
                        )
                else:
                    try:
                        y_vals = (
                            data_table["flux"][clip]
                            / data_table["flat"][clip]
                            / data_table["pscale"][clip]
                            / 1e-19
                        )
                        y_err = (
                            data_table["err"][clip]
                            / data_table["flat"][clip]
                            / data_table["pscale"][clip]
                            / 1e-19
                        )
                    except:
                        y_vals = (
                            data_table["flux"][clip] / data_table["flat"][clip] / 1e-19
                        )
                        y_err = (
                            data_table["err"][clip] / data_table["flat"][clip] / 1e-19
                        )

                    try:
                        update_errorbar(
                            self.plotted_components[dict_key][hdu.name],
                            data_table["wave"][clip],
                            y_vals,
                            yerr=y_err,
                        )
                    except:
                        self.plotted_components[dict_key][hdu.name] = (
                            self.fig_axes.errorbar(
                                data_table["wave"][clip],
                                y_vals,
                                yerr=y_err,
                                fmt="o",
                                markersize=3,
                                ecolor=colors.to_rgba(colours[hdu.name], 0.5),
                                c=colours[hdu.name],
                            )
                        )
                    ymax = np.nanmax([ymax, np.nanmax(y_vals)])

        if not templates:
            self.fig_axes.set_ylim(ymin=-0.05 * ymax, ymax=1.05 * ymax)

    def plot_MUSE_spec(
        self,
    ):
        if not self.cube_path.is_file():
            return
        if "MUSE_spec" in self.plotted_components.keys():
            for line in self.fig_axes.get_lines():
                if line == self.plotted_components["MUSE_spec"]:
                    line.remove()

        with pf.open(cube_path) as cube_hdul:
            cube_wcs = WCS(cube_hdul[1].header)

            wavelengths = (
                (np.arange(cube_hdul[1].header["NAXIS3"]) + 1.0)
                - cube_hdul[1].header["CRPIX3"]
            ) * cube_hdul[1].header["CD3_3"] + cube_hdul[1].header["CRVAL3"]
            MUSE_spec = self.cube_extract_spectra(
                cube_hdul[1].data,
                cube_wcs,
                self._root().tab_row[
                    self._root().config.get("catalogue", {}).get("ra", "X_WORLD")
                ],
                self._root().tab_row[
                    self._root().config.get("catalogue", {}).get("dec", "Y_WORLD")
                ],
                # radius=tab_row["r50_SE"][0],
            )

            (self.plotted_components["MUSE_spec"],) = self.fig_axes.plot(
                wavelengths,
                MUSE_spec
                / np.nanmedian(MUSE_spec)
                * np.nanmedian(self.fig_axes.get_ylim()),
                linewidth=0.5,
                c="k",
            )

    def cube_extract_spectra(
        self,
        data_cube,
        cube_wcs,
        ra,
        dec,
        radius=0.5,
        cube_error=None,
        kernel_sig=5,
    ):
        # temp_dir = (
        #     Path(self._root().config["files"]["temp_dir"]).expanduser().resolve()
        # )
        try:
            with pf.open(
                self.temp_dir
                / f"{ra[0]:.6f}_{dec[0]:.6f}_r{radius:.6f}_c{kernel_sig:.3f}.fits"
            ) as hdul:
                return hdul[0].data
        except Exception as e:
            print(e)
            try:
                ra.unit
                dec.unit
                sc = SkyCoord(
                    ra=ra,
                    dec=dec,
                )
            except:
                sc = SkyCoord(
                    ra=ra * u.deg,
                    dec=dec * u.deg,
                )
            try:
                radius.unit
                assert radius.unit is not None, ValueError
            except:
                print("failed")
                radius *= u.arcsec

            pix_c = np.hstack(sc.to_pixel(cube_wcs.celestial)[:])
            pix_r = radius / np.sqrt(cube_wcs.celestial.proj_plane_pixel_area()).to(
                radius.unit
            )

            aperture = CircularAperture(
                pix_c,
                pix_r.value,
            )

            spectrum = np.zeros(data_cube.shape[0])
            for i, cube_slice in tqdm(
                enumerate(data_cube[:]),
                desc="Extracting wavelength slice",
                total=len(spectrum),
            ):
                spectrum[i] = aperture_photometry(
                    cube_slice, aperture, error=cube_error
                )["aperture_sum"]

            kernel = Gaussian1DKernel(kernel_sig)
            spectrum = convolve(spectrum, kernel)

            new_hdul = pf.HDUList()
            new_hdul.append(
                pf.ImageHDU(data=spectrum, header=cube_wcs.spectral.to_header())
            )
            new_hdul.writeto(
                self.temp_dir
                / f"{ra[0]:.6f}_{dec[0]:.6f}_r{radius.value:.6f}_c{kernel_sig:.3f}.fits"
            )

            return spectrum

    def add_lines(
        self,
        line_type=None,
    ):
        if line_type is None:
            return
        xlims = self.fig_axes.get_xlim()
        for line_key, line_data in self._root().config["lines"][line_type].items():
            self.plotted_components[line_type][line_key] = self.fig.get_axes()[
                0
            ].axvline(
                line_data["centre"] * float(self.current_redshift.get()),
                c="0.7",
                alpha=0.7,
                linewidth=2,
            )

        self.fig_axes.set_xlim(xlims)
        self.pyplot_canvas.draw_idle()

    def update_lines(self, event=None):
        if type(event) == float:
            self.current_redshift.set(np.round(event, decimals=8))
        else:
            self.redshift_slider.set(float(self.current_redshift.get()))

        self._root().current_gal_data["estimated_redshift"] = float(
            self.current_redshift.get()
        )
        for line_type in ["emission", "absorption"]:
            try:
                for line_key, line_data in (
                    self._root().config["lines"][line_type].items()
                ):
                    current_line = self.plotted_components[line_type][line_key]
                    current_line.set_data(
                        [
                            line_data["centre"]
                            * (1 + float(self.current_redshift.get())),
                            line_data["centre"]
                            * (1 + float(self.current_redshift.get())),
                        ],
                        [0, 1],
                    )
            except:
                pass

        self.fig.canvas.draw_idle()

        self.redshift_plot.update_z_line()

        self.update()

    def reset_redshift(self):
        self.current_redshift.set(self.grizli_redshift)
        self.redshift_slider.set(self.grizli_redshift)
        self.update_lines()

    def change_components(self, event=None):
        if self.muse_checkbox.get():
            self.plot_MUSE_spec()
        elif "MUSE_spec" in self.plotted_components.keys():
            self.plotted_components["MUSE_spec"].remove()
            del self.plotted_components["MUSE_spec"]

        if self.grizli_checkbox.get():
            self.plot_grizli()
            view = True
        elif "grisms" in self.plotted_components.keys():
            view = False
        try:
            for v in self.plotted_components["grisms"].values():
                error_bar_visibility(v, view)
        except Exception as e:
            pass

        if self.grizli_temp_checkbox.get():
            self.plot_grizli(templates=True)
            view = True
        elif "grism_templates" in self.plotted_components.keys():
            view = False
        try:
            for v in self.plotted_components["grism_templates"].values():
                v.set_visible(view)
        except Exception as e:
            pass

        self.pyplot_canvas.draw_idle()

    def change_lines(self):
        if len(self.plotted_components["emission"]) == 0:
            self.add_lines(line_type="emission")
        # if self.emission_checkbox.get():
        for line in self.plotted_components["emission"].values():
            line.set_visible(self.emission_checkbox.get())
        if len(self.plotted_components["absorption"]) == 0:
            self.add_lines(line_type="absorption")
        for line in self.plotted_components["absorption"].values():
            line.set_visible(self.absorption_checkbox.get())

        self.all_plotted_lines = (
            self.plotted_components["emission"] | self.plotted_components["absorption"]
        )
        self.config_lines_data = (
            self._root().config["lines"]["emission"]
            | self._root().config["lines"]["absorption"]
        )

        self.update_lines()

        self.pyplot_canvas.draw_idle()

    def hover(self, event):
        if event.inaxes == self.fig_axes:
            for k, l in self.all_plotted_lines.items():
                if l.contains(event)[0] and l.get_visible():
                    self.custom_annotation.xy = [event.xdata, event.ydata]
                    self.custom_annotation.set_text(
                        self.config_lines_data[k]["tex_name"]
                    )
                    self.custom_annotation.set_visible(True)
                    self.fig.canvas.draw_idle()
                    return
        self.custom_annotation.set_visible(False)
        self.fig.canvas.draw_idle()


# based on https://gist.github.com/tacaswell/3144287
def zoom_factory(ax, base_scale=1.1):
    """
    Add ability to zoom with the scroll wheel.


    Parameters
    ----------
    ax : matplotlib axes object
        axis on which to implement scroll to zoom
    base_scale : float
        how much zoom on each tick of scroll wheel

    Returns
    -------
    disconnect_zoom : function
        call this to disconnect the scroll listener
    """

    def limits_to_range(lim):
        return lim[1] - lim[0]

    fig = ax.get_figure()  # get the figure of interest
    if hasattr(fig.canvas, "capture_scroll"):
        fig.canvas.capture_scroll = True
    has_toolbar = hasattr(fig.canvas, "toolbar") and fig.canvas.toolbar is not None
    if has_toolbar:
        toolbar = fig.canvas.toolbar
        toolbar.push_current()

    def zoom_fun(event):
        if event.inaxes is not ax:
            return
        # get the current x and y limits
        cur_xlim = ax.get_xlim()
        cur_ylim = ax.get_ylim()
        cur_yrange = limits_to_range(cur_ylim)
        cur_xrange = limits_to_range(cur_xlim)
        xdata = event.xdata  # get event x location
        ydata = event.ydata  # get event y location

        if event.button == "up":
            # deal with zoom in
            scale_factor = base_scale
        elif event.button == "down":
            # deal with zoom out
            scale_factor = 1 / base_scale
        else:
            # deal with something that should never happen
            scale_factor = 1
        # set new limits
        new_xlim = [
            xdata - (xdata - cur_xlim[0]) / scale_factor,
            xdata + (cur_xlim[1] - xdata) / scale_factor,
        ]
        new_ylim = [
            ydata - (ydata - cur_ylim[0]) / scale_factor,
            ydata + (cur_ylim[1] - ydata) / scale_factor,
        ]

        new_yrange = limits_to_range(new_ylim)
        new_xrange = limits_to_range(new_xlim)
        ax.set_xlim(new_xlim)
        ax.set_ylim(new_ylim)

        if has_toolbar:
            toolbar.push_current()
        ax.figure.canvas.draw_idle()  # force re-draw

    # attach the call back
    cid = fig.canvas.mpl_connect("scroll_event", zoom_fun)

    def disconnect_zoom():
        fig.canvas.mpl_disconnect(cid)

    # return the disconnect function
    return disconnect_zoom


class ImagesFrame(ctk.CTkFrame):
    def __init__(self, master, gal_id, **kwargs):
        super().__init__(master, **kwargs)

        self.gal_id = gal_id

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.update_seg_path()
        self.update_rgb_path()

        self.fig = Figure(
            constrained_layout=True,
            figsize=(10, 2),
        )
        self.pyplot_canvas = FigureCanvasTkAgg(
            figure=self.fig,
            master=self,
        )

        self.check_axes_colours()

        self.fig_axes = self.fig.subplots(
            1,
            5,
            sharey=True,
            # aspect="auto",
            # width_ratios=[1,shape_sci/shape_kernel],
            # width_ratios=[0.5,1]
        )
        for a in self.fig_axes:
            a.set_xticklabels("")
            a.set_yticklabels("")
            a.tick_params(axis="both", direction="in", top=True, right=True)
            # a.tick_params(axis="both", direction="in", left=False, bottom=False)

        self.default_cmap = colors.ListedColormap(["C1", "C0", "C2", "C3", "C4", "C5"])

        self.plotted_components = {}

        self.fig.canvas.draw_idle()

        self.fig.canvas.get_tk_widget().grid(row=0, column=0, sticky="news")

        if self._root().main_tabs.get() == "Spectrum":
            self.plot_seg_map()

    def check_axes_colours(self):
        self.fig.set_facecolor("none")
        self.fig.canvas.get_tk_widget().config(bg=self._root().bg_colour_name)

    def update_seg_path(self, pattern="*seg.fits"):
        self.seg_path = [
            *(
                Path(self._root().config["files"]["prep_dir"]).expanduser().resolve()
            ).glob(pattern)
        ]
        if len(self.seg_path) == 0:
            print("Segmentation map not found.")
            self.seg_path = None
        else:
            self.seg_path = self.seg_path[0]

    def update_rgb_path(self):
        self.rgb_paths = []
        for p in self._root().filter_names:
            rgb_path = [
                *(
                    Path(self._root().config["files"]["prep_dir"])
                    .expanduser()
                    .resolve()
                ).glob(f"*{p.lower()}*_dr[zc]_sci.fits")
            ]
            if len(rgb_path) == 0:
                print(f"{p} image not found.")
                self.rgb_paths.append(None)
            else:
                rgb_path.sort(reverse=True)
                self.rgb_paths.append(rgb_path[0])

    def plot_failed(self, ax, plot_name, text=None):
        for k in ["img", "marker", "text"]:
            try:
                self.plotted_components[f"{plot_name}_{k}"].set_visible(False)
            except Exception as e:
                pass
        if text is None:
            text = f"No data found\nfor {plot_name}."
        try:
            self.plotted_components[f"{plot_name}_failed"].set_text(text)
            self.plotted_components[f"{plot_name}_failed"].set_visible(True)
        except:
            self.plotted_components[f"{plot_name}_failed"] = ax.text(
                0.5,
                0.5,
                text,
                transform=ax.transAxes,
                ha="center",
                va="center",
                c=self._root().text_colour,
                visible=True,
            )

    def plot_images(self, border=5):
        plot_names = self._root().filter_names[::-1] + ["rgb", "seg"]

        # if self.seg_path is None and all(x is None for x in self.rgb_paths):
        #     print("No images to plot.")
        #     missing_data_text = self._root().filter_names[::-1] + [
        #         "No images found.",
        #         "Segmentation map\nnot found.",
        #     ]

        #     try:
        #         for v in self.plotted_components.values():
        #             v.set_visible(False)
        #     except:
        #         pass

        #     for a, n, t in zip(self.fig_axes, plot_names, missing_data_text):
        #         print(t)
        #         self.plotted_components[f"{n}_text"] = a.text(
        #             0.5,
        #             0.5,
        #             t,
        #             transform=a.transAxes,
        #             ha="center",
        #             va="center",
        #             c=self._root().text_colour,
        #             visible=True
        #         )
        #     self.pyplot_canvas.draw_idle()
        #     return

        try:
            with pf.open(self.seg_path) as hdul:
                seg_wcs = WCS(hdul[0].header)
                seg_data = hdul[0].data

                y_c, x_c = extract_pixel_ra_dec(
                    self._root().tab_row,
                    seg_wcs,
                    key_ra=self._root()
                    .config.get("catalogue", {})
                    .get("ra", "X_WORLD"),
                    key_dec=self._root()
                    .config.get("catalogue", {})
                    .get("dec", "Y_WORLD"),
                ).value

                location = np.where(seg_data == self._root().seg_id)
                width = np.nanmax(location[0]) - np.nanmin(location[0])
                height = np.nanmax(location[1]) - np.nanmin(location[1])

                if width > height:
                    w_d = 0
                    h_d = (width - height) / 2
                elif height > width:
                    h_d = 0
                    w_d = (height - width) / 2
                else:
                    w_d, h_d = 0, 0

                self.cutout_dimensions = [
                    int(
                        np.clip(
                            np.nanmin(location[0]) - border - w_d, 0, seg_data.shape[0]
                        )
                    ),
                    int(
                        np.clip(
                            np.nanmax(location[0]) + border + w_d, 0, seg_data.shape[0]
                        )
                    ),
                    int(
                        np.clip(
                            np.nanmin(location[1]) - border - h_d, 0, seg_data.shape[1]
                        )
                    ),
                    int(
                        np.clip(
                            np.nanmax(location[1]) + border + h_d, 0, seg_data.shape[1]
                        )
                    ),
                ]
                cutout = seg_data[
                    self.cutout_dimensions[0] : self.cutout_dimensions[1],
                    self.cutout_dimensions[2] : self.cutout_dimensions[3],
                ].astype(float)
                cutout[cutout == 0] = np.nan

                cutout_copy = cutout % 5 + 1
                cutout_copy[cutout == float(self._root().seg_id)] = 0

                try:
                    self.plotted_components["seg_img"].set_data(cutout_copy)
                    self.plotted_components["seg_img"].set_extent(
                        [0, cutout_copy.shape[0], 0, cutout_copy.shape[1]]
                    )
                    self.plotted_components["seg_img"].set_visible(True)
                except Exception as e:
                    self.plotted_components["seg_img"] = self.fig_axes[-1].imshow(
                        cutout_copy,
                        origin="lower",
                        cmap=self.default_cmap,
                        interpolation="nearest",
                        vmin=0,
                        vmax=5,
                        aspect="equal",
                        extent=[0, cutout_copy.shape[0], 0, cutout_copy.shape[1]],
                        visible=True,
                    )
                self.fig_axes[-1].set_xlim(xmax=cutout_copy.shape[0])
                self.fig_axes[-1].set_ylim(ymax=cutout_copy.shape[1])

                marker_xs = y_c - int(
                    np.clip(np.nanmin(location[1]) - border - h_d, 0, seg_data.shape[1])
                )
                marker_ys = x_c - int(
                    np.clip(np.nanmin(location[0]) - border - w_d, 0, seg_data.shape[0])
                )
                try:
                    self.plotted_components["seg_marker"].set_offsets(
                        (marker_xs, marker_ys)
                    )
                    self.plotted_components["seg_marker"].set_visible(True)
                except Exception as e:
                    self.plotted_components["seg_marker"] = self.fig_axes[-1].scatter(
                        marker_xs,
                        marker_ys,
                        marker="P",
                        c="k",
                        visible=True,
                    )
            if "seg_failed" in self.plotted_components.keys():
                self.plotted_components["seg_failed"].set_visible(False)
        except:
            self.plot_failed(
                ax=self.fig_axes[-1],
                plot_name="seg",
                text="Segmentation map\nnot found.",
            )

        try:
            self.rgb_data = np.empty(
                (
                    3,
                    self.cutout_dimensions[1] - self.cutout_dimensions[0],
                    self.cutout_dimensions[3] - self.cutout_dimensions[2],
                )
            )

            for i, v in enumerate(self.rgb_paths):
                with pf.open(v) as hdul:
                    try:
                        zp = hdul[0].header["ZP"]
                    except:
                        try:
                            # PASSAGE
                            zp = (
                                -2.5 * np.log10(hdul[0].header["PHOTFLAM"])
                                - 5 * np.log10(hdul[0].header["PHOTPLAM"])
                                - 2.408
                            )
                        except:
                            # Dawn JWST products
                            zp = 28.9
                    self.rgb_data[i] = hdul[0].data[
                        self.cutout_dimensions[0] : self.cutout_dimensions[1],
                        self.cutout_dimensions[2] : self.cutout_dimensions[3],
                    ] * 10 ** ((zp - 25) / 2.5)

            self.rgb_stretched = make_lupton_rgb(
                self.rgb_data[0],
                self.rgb_data[1],
                self.rgb_data[2],
                stretch=0.2,
            )
            try:
                self.plotted_components["rgb_img"].set_data(self.rgb_stretched)
                self.plotted_components["rgb_img"].set_extent(
                    [0, self.rgb_stretched.shape[0], 0, self.rgb_stretched.shape[1]]
                )
                self.plotted_components["rgb_img"].set_visible(True)
            except Exception as e:
                self.plotted_components["rgb_img"] = self.fig_axes[-2].imshow(
                    self.rgb_stretched,
                    origin="lower",
                    aspect="equal",
                    extent=[
                        0,
                        self.rgb_stretched.shape[0],
                        0,
                        self.rgb_stretched.shape[1],
                    ],
                    visible=True,
                )
            self.fig_axes[-2].set_xlim(xmax=self.rgb_stretched.shape[0])
            self.fig_axes[-2].set_ylim(ymax=self.rgb_stretched.shape[1])

            if "rgb_failed" in self.plotted_components.keys():
                self.plotted_components["rgb_failed"].set_visible(False)

            vmax = np.nanmax(
                [1.1 * np.percentile(self.rgb_data, 98), 5 * np.std(self.rgb_data)]
            )
            vmin = -0.1 * vmax
            interval = ManualInterval(vmin=vmin, vmax=vmax)
            for a, d, f in zip(
                self.fig_axes[:-2][::-1], self.rgb_data, self._root().filter_names
            ):
                try:
                    norm = ImageNormalize(
                        d,
                        interval=interval,
                        stretch=SqrtStretch(),
                    )
                    try:
                        self.plotted_components[f"{f}_img"].set_data(d)
                        self.plotted_components[f"{f}_img"].set_extent(
                            [0, d.shape[0], 0, d.shape[1]]
                        )
                        self.plotted_components[f"{f}_img"].set_norm(norm)
                    except:
                        self.plotted_components[f"{f}_img"] = a.imshow(
                            d,
                            origin="lower",
                            cmap="binary",
                            aspect="equal",
                            extent=[0, d.shape[0], 0, d.shape[1]],
                            norm=norm,
                        )

                    try:
                        self.plotted_components[f"{f}_text"].set_text(f)
                    except:
                        self.plotted_components[f"{f}_text"] = a.text(
                            0.05,
                            0.95,
                            f,
                            transform=a.transAxes,
                            ha="left",
                            va="top",
                            c="red",
                        )

                    if f"{f}_failed" in self.plotted_components.keys():
                        self.plotted_components[f"{f}_failed"].set_visible(False)
                except:
                    self.plot_failed(ax=a, plot_name=f)
        except Exception as e:
            self.plot_failed(ax=self.fig_axes[-2], plot_name="rgb")
            for a, f in zip(self.fig_axes[:3], plot_names[:3]):
                self.plot_failed(ax=a, plot_name=f)

        self.pyplot_canvas.draw_idle()
        # self.update()

    def update_images(self, force=False):
        self.check_axes_colours()
        if (
            self.gal_id != self._root().current_gal_id.get()
            or force
            or len(self.plotted_components) == 0
        ):
            self.gal_id = self._root().current_gal_id.get()
            self.plot_images()


def extract_pixel_radius(q_table, celestial_wcs, key="flux_radius"):
    # The assumption is that SExtractor radii are typically measured in pixel units
    radius = q_table[key]
    if hasattr(radius, "unit") and radius.unit != None:
        radius = radius.value * radius.unit  # Avoiding problems with columns
        if radius.unit == u.pix:
            pass
        elif u.get_physical_type(radius) == "dimensionless":
            radius *= u.pix
        elif u.get_physical_type(radius) == "angle":
            pixel_scale = (
                np.sqrt(celestial_wcs.proj_plane_pixel_area()).to(u.arcsec) / u.pix
            )
            radius /= pixel_scale
        else:
            raise ValueError(
                "The units of this radius cannot be automatically converted."
            )
    else:
        print("Radius has no unit, assuming pixels.")
        if hasattr(radius, "value"):
            radius = radius.value * u.pix
        else:
            radius = radius * u.pix

    return radius


def extract_pixel_ra_dec(q_table, celestial_wcs, key_ra="ra", key_dec="dec"):
    try:
        orig_ra = q_table[key_ra]
        orig_dec = q_table[key_dec]
    except:
        print(
            "No match found for supplied ra, dec keys. Performing automatic search instead."
        )
        lower_colnames = np.array([x.lower() for x in q_table.colnames])
        for r, d in [[key_ra, key_dec], ["ra", "dec"]]:
            possible_names = []
            for n in lower_colnames:
                if d.lower() in n:
                    possible_names.append(n)
            possible_names = sorted(possible_names, key=lambda x: (len(x), x))
            for n in possible_names:
                r_poss = n.replace(d.lower(), r.lower())
                if r_poss in lower_colnames:
                    orig_ra = q_table[
                        q_table.colnames[int((lower_colnames == r_poss).nonzero()[0])]
                    ]
                    orig_dec = q_table[
                        q_table.colnames[int((lower_colnames == n).nonzero()[0])]
                    ]
                    break
            else:
                continue
            break

    def check_deg(orig):
        if hasattr(orig, "unit") and orig.unit != None:
            new = orig.value * orig.unit  # Avoiding problems with columns
            if new.unit == u.pix:
                return new
            elif u.get_physical_type(new) == "dimensionless":
                new *= u.deg
            if u.get_physical_type(new) == "angle":
                new = new.to(u.deg)
        else:
            print("Coordinate has no unit, assuming degrees.")
            if hasattr(orig, "value"):
                new = orig.value * u.deg
            else:
                new = orig * u.deg
        return new

    new_ra, new_dec = check_deg(orig_ra), check_deg(orig_dec)
    if new_ra.unit == u.pix:
        return new_ra, new_dec

    sc = SkyCoord(new_ra, new_dec)
    pix_c = np.hstack(sc.to_pixel(celestial_wcs)[:]) * u.pix
    return pix_c


class RedshiftPlotFrame(ctk.CTkFrame):
    def __init__(self, master, gal_id, **kwargs):
        super().__init__(master, **kwargs)

        self.gal_id = gal_id

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.update_fits_path()

        self.fig = Figure(
            constrained_layout=True,
            # figsize=(10, 2),
        )
        self.pyplot_canvas = FigureCanvasTkAgg(
            figure=self.fig,
            master=self,
        )

        self.check_axes_colours()

        self.fig_axes = self.fig.subplots(
            1,
            1,
        )
        # for a in self.fig_axes:
        self.fig_axes.tick_params(axis="both", direction="in", top=True, right=True)
        self.fig_axes.set_xlabel(r"Redshift")
        self.fig_axes.set_ylabel(r"$\chi^2_{\rm red}$")

        self.plotted_components = {}

        self.fig.canvas.draw_idle()

        self.fig.canvas.get_tk_widget().grid(row=0, column=0, sticky="news")

        if self._root().main_tabs.get() == "Spectrum":
            self.update_z_grid()

    def check_axes_colours(self):
        self.fig.set_facecolor("none")
        self.fig.canvas.get_tk_widget().config(bg=self._root().bg_colour_name)

    def update_fits_path(self):
        pad = self._root().config.get("catalogue", {}).get("seg_id_length", 5)
        self.fits_path = [
            *(
                Path(self._root().config["files"]["extractions_dir"])
                .expanduser()
                .resolve()
            ).glob(f"**/*{self._root().seg_id:0>{pad}}.full.fits")
        ] + [
            *(
                Path(self._root().config["files"]["extractions_dir"])
                .expanduser()
                .resolve()
            ).glob(f"**/*{self._root().seg_id:0>{pad}}.maps.fits")
        ]
        if len(self.fits_path) == 0:
            print("Full extraction data not found.")
            self.fits_path = None
        else:
            self.fits_path = self.fits_path[0]

    def plot_z_grid(self):
        try:
            with pf.open(self.fits_path) as hdul_all:
                hdul = hdul_all["ZFIT_STACK"]
                try:
                    self.plotted_components["chi2_grid"].set_data(
                        hdul.data["zgrid"],
                        hdul.data["chi2"] / hdul.header["DOF"],
                    )
                except:
                    (self.plotted_components["chi2_grid"],) = self.fig_axes.plot(
                        hdul.data["zgrid"],
                        hdul.data["chi2"] / hdul.header["DOF"],
                    )
                self.fig_axes.relim()
                self.fig_axes.autoscale(axis="y")
                z_range = np.nanmax(hdul.data["zgrid"]) - np.nanmin(hdul.data["zgrid"])
                self.fig_axes.set_xlim(
                    [
                        np.nanmin(hdul.data["zgrid"]) - 0.05 * z_range,
                        np.nanmax(hdul.data["zgrid"]) + 0.05 * z_range,
                    ]
                )

            try:
                self.plotted_components[f"z_failed"].set_visible(False)
            except:
                pass
        except:
            try:
                self.plotted_components[f"z_failed"].set_visible(True)
            except:
                self.plotted_components[f"z_failed"] = self.fig_axes.text(
                    0.5,
                    0.5,
                    "No data.",
                    transform=self.fig_axes.transAxes,
                    ha="center",
                    va="center",
                    c=self._root().text_colour,
                    visible=True,
                )

        self.update_z_line()

    def update_z_line(self):
        if "z_line" not in self.plotted_components.keys():
            self.plotted_components["z_line"] = self.fig_axes.axvline(
                float(self.master.current_redshift.get()),
                c="0.7",
                alpha=0.7,
                linewidth=2,
            )
        else:
            self.plotted_components["z_line"].set_data(
                [
                    float(self.master.current_redshift.get()),
                    float(self.master.current_redshift.get()),
                ],
                [0, 1],
            )

        self.pyplot_canvas.draw_idle()
        # self.update()

    def update_z_grid(self, force=False):
        self.check_axes_colours()
        if (
            self.gal_id != self._root().current_gal_id.get()
            or force
            or len(self.plotted_components) == 0
        ):
            self.gal_id = self._root().current_gal_id.get()
            self.update_fits_path()
            self.plot_z_grid()



import re
from pathlib import Path

import astropy.units as u
import customtkinter as ctk
import matplotlib as mpl
import numpy as np
import tomlkit
from astropy.coordinates import SkyCoord
from astropy.table import QTable
from CTkMessagebox import CTkMessagebox
from tqdm import tqdm

from pygcg.tabs import BeamFrame, SpecFrame
from pygcg.utils import ValidateFloatVar, flatten_dict, fpe
from pygcg.windows import CommentsWindow, SearchWindow, SettingsWindow
from itertools import product


class GCG(ctk.CTk):
    def __init__(self, config_file=None):
        super().__init__()

        # Geometry
        self.geometry("1366x768")
        self.minsize(1280, 720)
        # self.attributes("-zoomed", True)
        self.title("GLASS-JWST Classification GUI")

        self.initialise_configuration(config_file)
        self.settings_window = None
        self.comments_window = None
        self.search_window = None

        # Key bindings
        self.protocol(
            "WM_DELETE_WINDOW",
            self.quit_gracefully,
        )

        self.bind("<Control-q>", self.quit_gracefully)
        self.bind("<Left>", self.prev_gal_button_callback)
        self.bind("<Right>", self.next_gal_button_callback)
        self.bind("<c>", self.gal_comments_button_callback)

        # configure grid system
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)

        # Setup bottom navigation buttons
        nav_frame = ctk.CTkFrame(self)
        nav_frame.grid(
            column=0,
            row=2,
            sticky="ew",
        )

        nav_frame.grid_columnconfigure((1, 2, 3, 5, 6), weight=1, uniform="settings")
        nav_frame.grid_columnconfigure((4), weight=0, uniform="frame")

        self.read_write_button = ctk.CTkSegmentedButton(
            nav_frame,
            values=["Read-only", "Write output"],
            selected_color="red",
            selected_hover_color="dark red",
            command=self.read_write_colour,
        )
        self.read_write_button.grid(
            row=0,
            column=1,
            padx=10,
            pady=10,
            sticky="ew",
        )
        self.read_write_button.set("Read-only")

        self.open_settings_button = ctk.CTkButton(
            nav_frame,
            text="Settings",
            command=self.open_settings_callback,
        )
        self.open_settings_button.grid(
            row=1,
            column=1,
            padx=10,
            pady=10,
            sticky="ew",
        )
        self.prev_gal_button = ctk.CTkButton(
            nav_frame,
            text="Previous",
            command=self.prev_gal_button_callback,
        )
        self.prev_gal_button.grid(
            row=0, column=0, padx=10, pady=10, rowspan=2, sticky="news"
        )
        self.next_gal_button = ctk.CTkButton(
            nav_frame,
            text="Next",
            command=self.next_gal_button_callback,
        )
        self.next_gal_button.grid(
            row=0, column=7, padx=10, pady=10, rowspan=2, sticky="news"
        )
        self.comments_button = ctk.CTkButton(
            nav_frame,
            text="Comments",
            command=self.gal_comments_button_callback,
        )
        self.comments_button.grid(
            row=0,
            column=6,
            padx=10,
            pady=10,
            sticky="ew",
        )

        self.current_gal_data = {}
        self.poss_extvers = []
        for g, p in product(
            [
                self.config["grisms"]["B"],
                self.config["grisms"]["G"],
                self.config["grisms"]["R"],
            ],
            [self.config["grisms"]["PA1"], self.config["grisms"]["PA2"]],
        ):
            self.poss_extvers.append(f"{g},{p}")
            self.current_gal_data[f"{g},{p}"] = {}

        self.warning_flag = True

        self.id_frame = ctk.CTkFrame(nav_frame, fg_color=self.bg_colour_name)
        self.id_frame.grid(
            column=2,
            row=0,
            columnspan=2,
            sticky="ew",
        )
        self.id_frame.grid_columnconfigure((0, 1), weight=1, uniform="blah")

        self.current_gal_id = ctk.StringVar(master=self)
        self.current_gal_coords = ctk.StringVar(master=self)
        gal_id_label = ctk.CTkLabel(
            self.id_frame,
            text="Current ID:",
        )
        gal_id_label.grid(
            row=0,
            column=0,
            padx=(10, 5),
            pady=10,
            sticky="e",
        )
        gal_coord_label = ctk.CTkLabel(
            nav_frame,
            text="Location:",
        )
        gal_coord_label.grid(
            row=0,
            column=4,
            padx=(10, 5),
            pady=10,
            sticky="e",
        )
        progress_label = ctk.CTkLabel(
            nav_frame,
            text="Progress:",
        )
        progress_label.grid(
            row=1,
            column=2,
            padx=(10, 5),
            pady=10,
            sticky="e",
        )
        self.progress_status = ctk.CTkLabel(
            nav_frame,
            text="",
        )
        self.progress_status.grid(
            row=1,
            column=3,
            padx=(5, 10),
            pady=10,
            sticky="w",
        )
        self.current_gal_entry = ctk.CTkEntry(
            self.id_frame,
            textvariable=self.current_gal_id,
        )
        self.current_gal_entry.bind(
            "<Return>",
            self.change_id_entry,
        )
        self.current_gal_entry.grid(
            row=0,
            column=1,
            padx=(5, 10),
            pady=10,
            sticky="w",
        )

        self.coord_entry = ctk.CTkEntry(
            nav_frame,
            textvariable=self.current_gal_coords,
        )
        self.coord_entry.bind(
            "<Return>",
            self.change_sky_coord,
        )
        self.coord_entry.grid(
            row=0,
            column=5,
            padx=(5, 10),
            pady=10,
            sticky="w",
        )
        self.search_button = ctk.CTkButton(
            nav_frame,
            text="Find Objects",
            command=self.search_button_callback,
        )
        self.search_button.grid(row=1, column=6, padx=10, pady=10, sticky="ew")

        self.gal_info_label = ctk.CTkLabel(
            nav_frame,
            text="",
        )
        self.gal_info_label.grid(
            row=1, column=4, columnspan=2, padx=(5, 10), pady=10, sticky="ew"
        )

        self.rescan_and_reload()

    def rescan_and_reload(self, skip=False):
        try:
            assert (
                len(self.config["files"]["extractions_dir"]) > 0
            ), "Extractions directory is not defined."

            try:
                self.cat = QTable.read(fpe(self.config["files"]["cat_path"]))
            except Exception as e:
                print("Catalogue could not be loaded from `cat_path' in config.")
                try:
                    self.cat = QTable.read(
                        [
                            *fpe(self.config["files"]["extractions_dir"]).glob(
                                "*ir.cat.fits"
                            )
                        ][0]
                    )
                except:
                    self.cat = None

            assert self.cat is not None, "No catalogue found."

            self.filter_names = [
                self.config.get("grisms", {}).get("R", "F200W"),
                self.config.get("grisms", {}).get("G", "F150W"),
                self.config.get("grisms", {}).get("B", "F115W"),
            ]
            self.PAs = [
                str(self.config.get("grisms", {}).get("PA1", 72.0)),
                str(self.config.get("grisms", {}).get("PA2", 341.0)),
            ]

            try:
                assert len(self.config["files"]["out_dir"]) > 0

                fpe(self.config["files"]["out_dir"]).mkdir(exist_ok=True, parents=True)
                self.out_dir = fpe(self.config["files"]["out_dir"])

                if len(self.config["files"]["temp_dir"]) > 0:
                    fpe(self.config["files"]["temp_dir"]).mkdir(
                        exist_ok=True, parents=True
                    )
                    self.temp_dir = fpe(self.config["files"]["temp_dir"])
                else:
                    self.temp_dir = self.out_dir / ".temp"
                    self.temp_dir.mkdir(exist_ok=True)

                out_name = self.config["files"].get("out_cat_name", "pyGCG_output.fits")
                if out_name == "":
                    out_name = "pyGCG_output.fits"
                self.out_cat_path = fpe(self.config["files"]["out_dir"]) / out_name
                self.read_write_button.configure(state="normal")
            except:
                print(
                    "Could not find or create output directory. Disabling write mode."
                )
                self.read_write_button.set("Read-only")
                self.read_write_button.configure(state="disabled")

            try:
                self.out_cat = QTable.read(self.out_cat_path)
            except:
                self.out_cat = QTable(
                    names=[
                        "ID",
                        "SEG_ID",
                        "RA",
                        "DEC",
                        f"{self.filter_names[2]},{self.PAs[0]}_QUALITY",
                        f"{self.filter_names[2]},{self.PAs[0]}_COVERAGE",
                        f"{self.filter_names[1]},{self.PAs[0]}_QUALITY",
                        f"{self.filter_names[1]},{self.PAs[0]}_COVERAGE",
                        f"{self.filter_names[0]},{self.PAs[0]}_QUALITY",
                        f"{self.filter_names[0]},{self.PAs[0]}_COVERAGE",
                        f"{self.filter_names[2]},{self.PAs[1]}_QUALITY",
                        f"{self.filter_names[2]},{self.PAs[1]}_COVERAGE",
                        f"{self.filter_names[1]},{self.PAs[1]}_QUALITY",
                        f"{self.filter_names[1]},{self.PAs[1]}_COVERAGE",
                        f"{self.filter_names[0]},{self.PAs[1]}_QUALITY",
                        f"{self.filter_names[0]},{self.PAs[1]}_COVERAGE",
                        "GRIZLI_REDSHIFT",
                        "ESTIMATED_REDSHIFT",
                        "UNRELIABLE_REDSHIFT",
                        "TENTATIVE_REDSHIFT",
                        "BAD_SEG_MAP",
                        "COMMENTS",
                    ],
                    dtype=[
                        str,
                        int,
                        float,
                        float,
                        str,
                        float,
                        str,
                        float,
                        str,
                        float,
                        str,
                        float,
                        str,
                        float,
                        str,
                        float,
                        float,
                        float,
                        bool,
                        bool,
                        bool,
                        str,
                    ],
                    units=[
                        None,
                        None,
                        "deg",
                        "deg",
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                    ],
                )

            for key, default in zip(
                ["id", "ra", "dec"], ["NUMBER", "X_WORLD", "Y_WORLD"]
            ):
                name = self.config.get("catalogue", {}).get(key, default)
                assert (
                    name in self.cat.colnames
                ), f'No <{key}> column with name "{name}" found in catalogue.'

            id_name = self.config.get("catalogue", {}).get("id", "NUMBER")
            self.id_col = np.array(self.cat[id_name].astype(str))

            seg_name = self.config.get("catalogue", {}).get(
                "seg_id", self.config.get("catalogue", {}).get("id", "NUMBER")
            )
            if not seg_name in self.cat.colnames:
                warn_seg = CTkMessagebox(
                    title="Invalid Configuration",
                    message=(
                        f'No <seg_id> column with name "{seg_name}" '
                        f'found in catalogue. Use <id> ("{id_name}") '
                        "instead, or cancel?"
                    ),
                    icon="warning",
                    option_1="Cancel",
                    option_2="Continue",
                    option_focus=2,
                )
                if warn_seg.get() == "Cancel":
                    return
                else:
                    seg_name = id_name

            self.seg_id_col = np.array(self.cat[seg_name].astype(int))

            if id_name != seg_name:
                self.id_frame.grid_columnconfigure((1, 3), weight=1, uniform="entry")

                self.seg_id_label = ctk.CTkLabel(
                    self.id_frame,
                    text="Segmentation ID:",
                )
                self.seg_id_label.grid(
                    row=0,
                    column=2,
                    padx=(10, 5),
                    pady=10,
                    sticky="e",
                )
                self.current_seg_id = ValidateFloatVar(self)
                self.current_seg_entry = ctk.CTkEntry(
                    self.id_frame,
                    textvariable=self.current_seg_id,
                )
                self.current_seg_entry.bind(
                    "<Return>",
                    self.change_seg_entry,
                )
                self.current_seg_entry.grid(
                    row=0,
                    column=3,
                    padx=(5, 10),
                    pady=10,
                    sticky="w",
                )
            else:
                try:
                    self.current_seg_entry.destroy()
                    del self.current_seg_id
                    self.seg_id_label.destroy()
                    self.id_frame.grid_columnconfigure((0, 1), weight=1, uniform="blah")
                except Exception as e:
                    pass

            # Segmentation map ids must be a unique identifier!
            # If you're reading this comment, something has gone horribly wrong
            _, unique_idx = np.unique(self.seg_id_col, return_index=True)
            unique_idx = np.sort(unique_idx)
            self.seg_id_col = self.seg_id_col[unique_idx]
            self.id_col = self.id_col[unique_idx]
            self.cat = self.cat[unique_idx]
            dir_to_chk = fpe(self.config["files"]["extractions_dir"])

            id_idx_list = []

            pad = self.config.get("catalogue", {}).get("seg_id_length", 5)

            stack_ids = [
                s.stem[-6 - pad : -6] for s in dir_to_chk.glob("**/*.stack.fits")
            ] + [
                s.stem[-7 - pad : -7] for s in dir_to_chk.glob("**/*.spec2D.fits")
            ]
            oned_ids = [
                s.stem[-3 - pad : -3] for s in dir_to_chk.glob("**/*.1D.fits")
            ] + [s.stem[-7 - pad : -7] for s in dir_to_chk.glob("**/*.spec1D.fits")]

            for i, (n, s) in tqdm(
                enumerate(zip(self.id_col, self.seg_id_col)),
                desc="Scanning directory for objects in catalogue",
                total=len(self.id_col),
            ):
                if self.config["files"].get("skip_existing", True) or skip:
                    if s in self.out_cat["SEG_ID"]:
                        continue
                if (
                    f"{s:0>{pad}}" in oned_ids
                    and f"{s:0>{pad}}" in stack_ids
                ):
                    id_idx_list.append(i)

            assert len(id_idx_list) > 0, (
                f"No matches found in the extractions directory for the "
                f"{len(self.id_col)} objects in the catalogue."
            )

            self.id_col = self.id_col[id_idx_list]
            self.seg_id_col = self.seg_id_col[id_idx_list]
            self.cat = self.cat[id_idx_list]
            try:
                self.sky_coords = SkyCoord(
                    self.cat[self.config.get("catalogue", {}).get("ra", "X_WORLD")],
                    self.cat[self.config.get("catalogue", {}).get("dec", "Y_WORLD")],
                )
            except:
                self.sky_coords = SkyCoord(
                    self.cat[self.config.get("catalogue", {}).get("ra", "X_WORLD")],
                    self.cat[self.config.get("catalogue", {}).get("dec", "Y_WORLD")],
                    unit="deg"
                )

            assert len(self.id_col) > 0

            self.current_gal_id.set(self.id_col[0])
            self.tab_row = self.cat[0]
            self.seg_id = self.seg_id_col[0]
            if hasattr(self, "current_seg_id"):
                self.current_seg_id.set(self.seg_id)

            self.set_current_data()

            self.generate_tabs()
        except Exception as e:
            error = CTkMessagebox(
                title="Error",
                message=e,
                icon="cancel",
                option_focus=1,
            )
            if error.get() == "OK":
                self.generate_splash()

    def generate_splash(self):
        self.splash_frame = ctk.CTkFrame(self)
        self.splash_frame.grid(row=0, column=0, rowspan=2, sticky="news")
        self.splash_frame.columnconfigure(0, weight=1)
        self.splash_frame.rowconfigure(0, weight=1)
        main_label = ctk.CTkLabel(
            self.splash_frame,
            text=(
                "No objects found. Check the supplied directories, \n"
                "or rescan the current directory in the settings menu."
            ),
            font=ctk.CTkFont(family="", size=20),
        )
        main_label.grid(row=0, column=0, sticky="news")

    def generate_tabs(self):
        if hasattr(self, "splash_frame"):
            self.splash_frame.destroy()
            del self.splash_frame

        self.plot_options = {
            "cmap": "plasma",
            "stretch": "Square root",
            "limits": "grizli default",
        }

        self.tab_names = [
            f"Orientation 1: {self.PAs[0]} deg",
            f"Orientation 2: {self.PAs[1]} deg",
            f"Spectrum",
        ]
        self.object_progress = {}
        for n in self.tab_names:
            self.object_progress[n] = False
        self.update_progress()

        self.main_tabs = MyTabView(
            master=self,
            tab_names=self.tab_names,
            command=self.main_tabs_update,
        )
        self.main_tabs.grid(row=1, column=0, padx=10, pady=0, sticky="news")
        self.main_tabs._segmented_button.grid(sticky="ew")

        self.full_spec_frame = SpecFrame(
            self.main_tabs.tab(self.tab_names[2]), self.current_gal_id.get()
        )
        self.full_spec_frame.pack(fill="both", expand=1)

        self.beam_frame_1 = BeamFrame(
            self.main_tabs.tab(self.tab_names[0]),
            self.current_gal_id.get(),
            self.PAs[0],
        )
        self.beam_frame_1.pack(fill="both", expand=1)
        self.beam_frame_2 = BeamFrame(
            self.main_tabs.tab(self.tab_names[1]),
            self.current_gal_id.get(),
            self.PAs[1],
        )
        self.beam_frame_2.pack(fill="both", expand=1)

        self.quality_key_map = np.array(
            [["q", "w", "e", "r"], ["u", "i", "o", "p"], ["a", "s", "d", "f"]]
        )
        for l in self.quality_key_map.flatten():
            self.bind(f"{l}", self.select_quality_menu)

        self.object_progress[self.main_tabs.get()] = True
        self.update_progress()

    def select_quality_menu(self, event=None):
        if event != None and event.widget.winfo_class() == ("Entry" or "Textbox"):
            return

        if self.beam_frame_1.winfo_viewable():
            widg = self.beam_frame_1
        else:
            widg = self.beam_frame_2
        widg.beam_single_PA_frame.quality_frame.keypress_select(
            event, self.quality_key_map
        )

    def initialise_configuration(self, config_file=None):
        try:
            if config_file is not None:
                self.config_file_path = fpe(config_file)
            else:
                self.config_file_path = "./config.toml"
            with open(self.config_file_path, "rt") as fp:
                self.config = tomlkit.load(fp)
            self.write_config()
        except Exception as e:
            print(e)
            print(
                "No valid config file supplied. "
                "Creating config.toml in the current working directory."
            )
            example_path = Path(__file__).parent / "example_config.toml"
            with open(example_path, "rt") as fp:
                self.config = tomlkit.load(fp)
            self.config_file_path = fpe(Path.cwd() / "config.toml")
            with open(
                self.config_file_path,
                mode="wt",
                encoding="utf-8",
            ) as fp:
                tomlkit.dump(self.config, fp)

        ctk.set_appearance_mode(self.config["appearance"]["appearance_mode"].lower())
        ctk.set_default_color_theme(self.config["appearance"]["theme"].lower())
        self.change_plot_colours()

    def change_plot_colours(self):
        if ctk.get_appearance_mode() == "Dark":
            mode = 1
        else:
            mode = 0
        self.bg_colour_name = ctk.ThemeManager.theme["CTkFrame"]["fg_color"][mode]
        self.bg_colour = [
            a / 65535
            for a in self.winfo_rgb(
                ctk.ThemeManager.theme["CTkFrame"]["fg_color"][mode]
            )
        ]
        self.text_colour = [
            a / 65535
            for a in self.winfo_rgb(
                ctk.ThemeManager.theme["CTkLabel"]["text_color"][mode]
            )
        ]

        mpl.rcParams["text.color"] = self.text_colour
        mpl.rcParams["axes.labelcolor"] = self.text_colour
        mpl.rcParams["xtick.color"] = self.text_colour
        mpl.rcParams["ytick.color"] = self.text_colour
        mpl.rcParams["axes.edgecolor"] = self.text_colour
        mpl.rcParams["axes.facecolor"] = "white" if mode == 0 else "k"

    def write_config(self):
        try:
            files = self.config["files"]
        except:
            files_tab = tomlkit.table()
            self.config.add("files", files_tab)
            files = self.config["files"]

        for f in ["extractions_dir"]:
            try:
                files[f]
            except:
                files.add(f, "")

        # Appearance
        try:
            appearance = self.config["appearance"]
        except:
            appearance_tab = tomlkit.table()
            self.config.add("appearance", appearance_tab)
            appearance = self.config["appearance"]

        try:
            appearance["appearance_mode"]
        except:
            appearance.add("appearance_mode", "system")
            appearance["appearance_mode"].comment("System (default), light, or dark.")

        try:
            appearance["theme"]
        except:
            appearance.add("theme", "blue")
            appearance["theme"].comment(
                "Blue (default), dark-blue, or green. The CustomTKinter color theme. "
                "This can also point to the location of a custom .json file describing"
                " the desired theme."
            )

        # Lines
        try:
            lines = self.config["lines"]
        except:
            lines_tab = tomlkit.table()
            lines_tab.add(
                tomlkit.comment(
                    "These tables define the lines shown in the redshift tab."
                )
            )
            lines_tab.add(tomlkit.nl())
            self.config.add(tomlkit.nl())
            self.config.add("lines", lines_tab)
            lines = self.config["lines"]

        try:
            emission = lines["emission"]
        except:
            lines.add("emission", tomlkit.table().indent(4))
            emission = lines["emission"]
            emission.add(tomlkit.comment("These are the emission lines."))
            emission.add(tomlkit.nl())

        em_lines = {
            "Lyman_alpha": {
                "tex_name": r"Ly$\alpha$",
                "centre": 1215.24,
            },
            "H_alpha": {
                "tex_name": r"H$\alpha$",
                "centre": 6564.61,
            },
        }

        for line_name, line_data in em_lines.items():
            try:
                emission[line_name]
                for key in line_data.keys():
                    emission[line_name][key]
            except:
                emission.add(line_name, tomlkit.table().indent(4))
                for key, value in line_data.items():
                    emission[line_name].add(key, value)
                emission.add(tomlkit.nl())

        try:
            absorption = lines["absorption"]
        except:
            lines.add("absorption", tomlkit.table().indent(4))
            absorption = lines["absorption"]
            absorption.add(tomlkit.comment("These are the absorption lines."))
            absorption.add(tomlkit.nl())

        with open(
            fpe(self.config_file_path),
            mode="wt",
            encoding="utf-8",
        ) as fp:
            tomlkit.dump(self.config, fp)

        return self.config

    def update_progress(self):
        num = np.sum([*self.object_progress.values()])
        blocks = num * 4 * "\u2588"
        self.progress_status.configure(text=f"|{blocks:\u2591<12}| {num}/3")

    def open_settings_callback(self, event=None):
        if self.settings_window is None or not self.settings_window.winfo_exists():
            self.settings_window = SettingsWindow(self)
        else:
            self.settings_window.focus()

    def gal_comments_button_callback(self, event=None):
        if event != None and event.widget.winfo_class() == ("Entry" or "Textbox"):
            return
        if self.comments_window is None or not self.comments_window.winfo_exists():
            self.comments_window = CommentsWindow(self)
        else:
            self.comments_window.focus()

    def prev_gal_button_callback(self, event=None):
        if not hasattr(self, "main_tabs"):
            return
        if self.warning_flag:
            self.check_read_write()
            self.warning_flag = False
            self.focus_force()

        if event != None and event.widget.winfo_class() == ("Entry" or "Textbox"):
            return
        current_tab = self.main_tabs.get()
        if current_tab == self.tab_names[2]:
            self.main_tabs.set(self.tab_names[1])
            self.main_tabs_update()
        elif current_tab == self.tab_names[1]:
            self.main_tabs.set(self.tab_names[0])
            self.main_tabs_update()
        elif current_tab == self.tab_names[0]:
            self.save_current_object()
            current_gal_idx = (self.id_col == self.current_gal_id.get()).nonzero()[0]
            self.current_gal_id.set(
                self.id_col[(current_gal_idx - 1) % len(self.id_col)][0]
            )
            self.main_tabs.set(self.tab_names[2])
            self.change_gal_id()
        self.object_progress[self.main_tabs.get()] = True
        self.update_progress()

    def next_gal_button_callback(self, event=None):
        if not hasattr(self, "main_tabs"):
            return
        if self.warning_flag:
            self.check_read_write()
            self.warning_flag = False
            self.focus_force()

        if event != None and event.widget.winfo_class() == ("Entry" or "Textbox"):
            return
        current_tab = self.main_tabs.get()
        if current_tab == self.tab_names[0]:
            self.main_tabs.set(self.tab_names[1])
            self.main_tabs_update()
        elif current_tab == self.tab_names[1]:
            self.main_tabs.set(self.tab_names[2])
            self.main_tabs_update()
        elif current_tab == self.tab_names[2]:
            self.save_current_object()
            if self.current_gal_id.get() == self.id_col[-1]:
                match self.check_end_objects():
                    case "continue":
                        pass
                    case "cancel":
                        return
                    case "reload":
                        self.rescan_and_reload(skip=True)
                        return
                    case "quit":
                        self.quit_gracefully()
            current_gal_idx = (self.id_col == self.current_gal_id.get()).nonzero()[0]
            self.current_gal_id.set(
                self.id_col[(current_gal_idx + 1) % len(self.id_col)][0]
            )
            self.main_tabs.set(self.tab_names[0])
            self.change_gal_id()
        self.object_progress[self.main_tabs.get()] = True
        self.update_progress()

    def check_read_write(self):
        mid_message = (
            "Classifications will not be saved."
            if self.read_write_button.get().lower() == "read-only"
            else (
                f"Classifications will be written to "
                f"{self.out_cat_path.name} in the output directory."
            )
        )
        check_overwrite = CTkMessagebox(
            title="Read/Write Mode",
            message=(
                f"This program is currently set to "
                f"{self.read_write_button.get().lower()}."
                f"{mid_message} If this is the expected behaviour, please continue."
                "\n\n(This message will not be shown again.)"
            ),
            icon="info",
            option_1="OK",
            option_focus=1,
        )

        if check_overwrite.get() == "OK":
            return

    def check_end_objects(self):
        num_classified = len(
            [a for a in self.seg_id_col if a in self.out_cat["SEG_ID"].value]
        )

        if num_classified < len(self.seg_id_col):
            check_end = CTkMessagebox(
                title="End of Input Catalogue",
                message=(
                    "You have reached the end of the current object list. However, "
                    f"only {num_classified}/{len(self.seg_id_col)} objects have been "
                    "classified. Continue viewing the existing objects, or reload "
                    "only the unclassified objects?\n\n"
                    "(If you want to skip classified objects on program load, set "
                    "`skip_existing = True' in the config file.)"
                ),
                icon="info",
                option_3="Continue",
                option_2="Reload",
                option_1="Cancel",
                option_focus=3,
                width=500,
            )

        else:
            check_end = CTkMessagebox(
                title="Read/Write Mode",
                message=(
                    "You have reached the end of the current object list, and all "
                    "objects have been classified. Continue viewing objects, or quit?"
                ),
                icon="info",
                option_2="Continue",
                option_1="Quit",
                option_focus=2,
            )

        out = check_end.get()
        try:
            out = out.lower()
        except:
            out = "cancel"

        self.focus_force()
        return out

    def save_current_object(self, event=None):
        ### This is where the logic for loading/updating the tables will go
        flattened_data = flatten_dict(self.current_gal_data)

        if (
            len(flattened_data) == 22
            and self.read_write_button.get() == "Write output"
            and np.sum([*self.object_progress.values()]) == 3
        ):
            if flattened_data["SEG_ID"] in self.out_cat["SEG_ID"]:
                warn_overwrite = CTkMessagebox(
                    title="Object already classified!",
                    message=(
                        "The data products for this object have already been classified. "
                        "Overwrite existing classification?\n"
                        "(This operation cannot be undone.)"
                    ),
                    icon="warning",
                    option_1="Cancel",
                    option_2="Overwrite",
                    option_focus=2,
                )
                if warn_overwrite.get() == "Cancel":
                    self.focus_force()
                    return
                else:
                    self.out_cat.remove_rows(
                        (flattened_data["SEG_ID"] == self.out_cat["SEG_ID"]).nonzero()[
                            0
                        ]
                    )
                    self.focus_force()
            self.out_cat.add_row(flattened_data)
            self.out_cat.write(self.out_cat_path, overwrite=True)

    def change_gal_id(self, event=None):
        for n in self.tab_names:
            self.object_progress[n] = False
        self.update_progress()
        self.tab_row = self.cat[self.id_col == self.current_gal_id.get()][0]
        self.seg_id = self.seg_id_col[self.id_col == self.current_gal_id.get()][0]
        if hasattr(self, "current_seg_id"):
            self.current_seg_id.set(self.seg_id)

        self.set_current_data()

        self.main_tabs_update()
        self.focus_force()

    def set_current_data(self):
        self.current_gal_data = {}
        self.current_gal_data["id"] = self.current_gal_id.get()
        self.current_gal_data["seg_id"] = self.seg_id
        self.current_gal_data["ra"] = self.tab_row[
            self.config.get("catalogue", {}).get("ra", "X_WORLD")
        ]
        self.current_gal_data["dec"] = self.tab_row[
            self.config.get("catalogue", {}).get("dec", "Y_WORLD")
        ]
        self.current_gal_data["comments"] = ""

        self.current_gal_coords.set(
            self.sky_coords[self.seg_id_col == self.seg_id].to_string(
                "decimal", precision=6
            )[0]
        )

        try:
            mag_name = self.config.get("catalogue", {}).get("mag", "MAG_AUTO")
            mag_val = self.tab_row[mag_name].value
            mag_text = f"mag\u2090\u1d64\u209c\u2092 = {mag_val:.2f}    "
        except:
            mag_text = ""
        try:
            rad_name = self.config.get("catalogue", {}).get("radius", "KRON_RCIRC")
            plate_scale = self.config.get("catalogue", {}).get("plate_scale", None)
            if plate_scale is None:
                rad_val = f"{self.tab_row[rad_name]:.0f}"
            else:
                rad_val = float(plate_scale) * self.tab_row[rad_name].value * u.arcsec
                rad_val = f"{rad_val:.2f}"
            rad_text = f"r\u2096\u1d63\u2092\u2099 = {rad_val}"
        except Exception as e:
            rad_text = ""
        self.gal_info_label.configure(
            text=(f"{mag_text}" f"{rad_text}"),
        )

        if self.seg_id in self.out_cat["SEG_ID"]:
            out_row = self.out_cat[self.out_cat["SEG_ID"] == self.seg_id][0]
            if not hasattr(out_row["COMMENTS"], "mask"):
                self.current_gal_data["comments"] = out_row["COMMENTS"]

            for g in self.filter_names:
                for p in self.PAs:
                    self.current_gal_data[f"{g},{p}"] = {
                        "quality": out_row[f"{g},{p}_QUALITY"]
                    }
            self.current_gal_data["grizli_redshift"] = out_row["GRIZLI_REDSHIFT"]
            self.current_gal_data["estimated_redshift"] = out_row["ESTIMATED_REDSHIFT"]
            self.current_gal_data["unreliable_redshift"] = out_row[
                "UNRELIABLE_REDSHIFT"
            ]
            self.current_gal_data["tentative_redshift"] = out_row[
                "TENTATIVE_REDSHIFT"
            ]
            self.current_gal_data["bad_seg_map"] = out_row["BAD_SEG_MAP"]

        for gp in self.poss_extvers:
            self.current_gal_data[gp] = {}

    def change_sky_coord(self, event=None):
        new_coord = None
        try:
            new_coord = SkyCoord(self.current_gal_coords.get())
        except ValueError:
            try:
                parts = re.split(r"\s*[,|;|\s]\s*", self.current_gal_coords.get())
                if len(parts) == 2:
                    new_coord = SkyCoord(
                        float(parts[0]) * u.deg, float(parts[1]) * u.deg
                    )
            except Exception as e:
                print(e)
                pass
        except Exception as e:
            print(e)
            pass

        if new_coord is None:
            error = CTkMessagebox(
                title="Error",
                message="Could not parse input as on-sky coordinates.",
                icon="cancel",
                option_focus=1,
            )
            if error.get() == "OK":
                self.focus_force()
                return

        sky_match_idx, dist, _ = new_coord.match_to_catalog_sky(self.sky_coords)

        print(
            f"Closest match: ID {self.id_col[sky_match_idx]}, "
            f"on-sky distance {dist[0].to(u.arcsec)}."
        )

        self.current_gal_id.set(self.id_col[sky_match_idx])

        self.focus_force()
        self.save_current_object()
        self.change_gal_id()

    def change_id_entry(self, event=None):
        self.save_current_object()

        try:
            new_id = self.id_col[self.id_col == self.current_gal_entry.get()][0]
            self.current_gal_id.set(new_id)
            self.change_gal_id()
        except:
            error = CTkMessagebox(
                title="Error",
                message="Input is not a valid ID.",
                icon="cancel",
                option_focus=1,
            )
            if error.get() == "OK":
                self.id_col[self.seg_id_col == self.seg_id][0]
                self.current_gal_id.set(self.id_col[self.seg_id_col == self.seg_id][0])
                self.focus_force()
                return

    def change_seg_entry(self, event=None):
        self.save_current_object()

        try:
            new_id = self.id_col[self.seg_id_col == int(self.current_seg_entry.get())][
                0
            ]
            self.current_gal_id.set(new_id)
            self.change_gal_id()
        except:
            error = CTkMessagebox(
                title="Error",
                message="Input is not a valid segmentation map ID.",
                icon="cancel",
                option_focus=1,
            )
            if error.get() == "OK":
                self.current_seg_id.set(self.seg_id)
                self.focus_force()
                return

    def search_button_callback(self, event=None):
        if event != None and event.widget.winfo_class() == ("Entry" or "Textbox"):
            return
        if self.search_window is None or not self.search_window.winfo_exists():
            self.search_window = SearchWindow(self)
        else:
            self.search_window.focus()

    def read_write_colour(self, event=None):
        if self.read_write_button.get() == "Read-only":
            self.read_write_button.configure(
                selected_color="red", selected_hover_color="dark red"
            )
        if self.read_write_button.get() == "Write output":
            self.read_write_button.configure(
                selected_color="green", selected_hover_color="dark green"
            )

    def main_tabs_update(self):
        self.object_progress[self.main_tabs.get()] = True
        self.update_progress()
        if self.main_tabs.get() == self.tab_names[2]:
            self.full_spec_frame.update_plot()
        if self.main_tabs.get() == self.tab_names[0]:
            self.beam_frame_1.update_grid()
        if self.main_tabs.get() == self.tab_names[1]:
            self.beam_frame_2.update_grid()

    def quit_gracefully(self, event=None):
        self.write_config()
        self.quit()


class MyTabView(ctk.CTkTabview):
    def __init__(self, master, tab_names, expose_bind_fns=None, **kwargs):
        super().__init__(master, **kwargs)

        # create tabs
        for i, name in enumerate(tab_names):
            self.add(name)
            try:
                self.tab(name).bind("<<TabChanged>>", expose_bind_fns[i])
            except:
                pass


def run_app(**kwargs):
    app = GCG(**kwargs)
    app.mainloop()
    app.withdraw()



import tkinter as tk
from pathlib import Path

import astropy.io.fits as pf
import astropy.units as u
import customtkinter as ctk
import matplotlib as mpl
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table
from astropy.visualization import (
    ImageNormalize,
    LinearStretch,
    LogStretch,
    ManualInterval,
    MinMaxInterval,
    PercentileInterval,
    SqrtStretch,
)
from astropy.wcs import WCS
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
from tqdm import tqdm


class BeamFrame(ctk.CTkFrame):
    def __init__(self, master, gal_id, PA, **kwargs):
        super().__init__(master, **kwargs)

        self.settings_frame = ctk.CTkFrame(self)
        self.settings_frame.grid(
            row=0,
            column=0,
            columnspan=3,
            sticky="ew",
        )
        self.PA = PA

        cmap_label = ctk.CTkLabel(self.settings_frame, text="Colourmap:")
        cmap_label.grid(row=0, column=2, padx=(20, 5), pady=10, sticky="e")
        self.cmap_menu = ctk.CTkOptionMenu(
            self.settings_frame,
            values=[
                "plasma",
                "plasma_r",
                "viridis",
                "viridis_r",
                "jet",
                "binary",
                "binary_r",
            ],
            command=self.change_cmap,
        )
        self.cmap_menu.grid(row=0, column=3, padx=(5, 20), pady=10, sticky="w")
        self.cmap_menu.set(self._root().plot_options["cmap"])

        stretch_label = ctk.CTkLabel(self.settings_frame, text="Image stretch:")
        stretch_label.grid(row=0, column=4, padx=(20, 5), pady=10, sticky="e")
        self.stretch_menu = ctk.CTkOptionMenu(
            self.settings_frame,
            values=["Linear", "Square root", "Logarithmic"],
            command=self.change_stretch,
        )
        self.stretch_menu.set(self._root().plot_options["cmap"])
        self.stretch_menu.grid(row=0, column=5, padx=(5, 20), pady=10, sticky="w")

        limits_label = ctk.CTkLabel(self.settings_frame, text="Colourmap limits:")
        limits_label.grid(row=0, column=6, padx=(20, 5), pady=10, sticky="e")
        self.limits_menu = ctk.CTkOptionMenu(
            self.settings_frame,
            values=[
                "grizli default",
                "Min-max",
                "99.9%",
                "99.5%",
                "99%",
                "98%",
                "95%",
                "90%",
            ],
            command=self.change_limits,
        )
        self.limits_menu.set(self._root().plot_options["limits"])
        self.limits_menu.grid(row=0, column=7, padx=(5, 20), pady=10, sticky="w")

        self.gal_id = gal_id
        try:
            pad = self._root().config.get("catalogue", {}).get("seg_id_length", 5)
            self.file_path = [
                *(
                    Path(self._root().config["files"]["extractions_dir"])
                    .expanduser()
                    .resolve()
                ).glob(f"**/*{self._root().seg_id:0>{pad}}.stack.fits")
            ] + [
                *(
                    Path(self._root().config["files"]["extractions_dir"])
                    .expanduser()
                    .resolve()
                ).glob(f"**/*{self._root().seg_id:0>{pad}}.spec2D.fits")
            ]
            self.file_path = self.file_path[0]
        except:
            self.file_path = None

        self.generate_grid()

    def change_cmap(self, event=None):
        self._root().plot_options["cmap"] = event
        self.update_grid(force_update=True)

    def change_stretch(self, event=None):
        self._root().plot_options["stretch"] = event
        self.update_grid(force_update=True)

    def change_limits(self, event=None):
        self._root().plot_options["limits"] = event
        self.update_grid(force_update=True)

    def update_grid(self, force_update=False):
        if self.gal_id == self._root().current_gal_id.get() and not force_update:
            self.beam_single_PA_frame.quality_frame.save_current()
            for k, v in self.beam_single_PA_frame.coverage.items():
                try:
                    self._root().current_gal_data[k]["coverage"] = v
                except:
                    pass
            pass
        else:
            self.gal_id = self._root().current_gal_id.get()
            pad = self._root().config.get("catalogue", {}).get("seg_id_length", 5)
            self.file_path = [
                *(
                    Path(self._root().config["files"]["extractions_dir"])
                    .expanduser()
                    .resolve()
                ).glob(f"**/*{self._root().seg_id:0>{pad}}.stack.fits")
            ] + [
                *(
                    Path(self._root().config["files"]["extractions_dir"])
                    .expanduser()
                    .resolve()
                ).glob(f"**/*{self._root().seg_id:0>{pad}}.spec2D.fits")
            ]
            self.file_path = self.file_path[0]

            extver_list = [s for s in self._root().poss_extvers if self.PA in s]
            self.beam_single_PA_frame.update_plots(extvers=extver_list)

            self.update()

    def generate_grid(self):
        with pf.open(self.file_path) as hdul:
            header = hdul[0].header
            n_grism = header["NGRISM"]
            n_pa = np.nanmax(
                [header[f"N{header[f'GRISM{n:0>3}']}"] for n in range(1, n_grism + 1)]
            )
            extver_list = [s for s in self._root().poss_extvers if self.PA in s]

            self.beam_single_PA_frame = SinglePABeamFrame(self, extvers= extver_list)
            self.beam_single_PA_frame.grid(row=1, column=0, sticky="news")
            self.grid_rowconfigure(1, weight=1)
            self.grid_columnconfigure(0, weight=1)


class SinglePABeamFrame(ctk.CTkFrame):
    def __init__(
        self,
        master,
        extvers,
        **kwargs,
    ):
        super().__init__(master, **kwargs)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.pad_frame = ctk.CTkFrame(self)  # , fg_color="red")
        self.pad_frame.grid(row=0, column=0, sticky="news")
        self.canvas_frame = ctk.CTkFrame(self.pad_frame)  # , fg_color="blue")
        self.canvas_frame.grid(row=0, column=0, sticky="news")
        self.canvas_frame.rowconfigure(0, weight=1)
        self.canvas_frame.columnconfigure(0, weight=1)

        self.fig = Figure(
            constrained_layout=True,
        )
        self.pyplot_canvas = FigureCanvasTkAgg(
            figure=self.fig,
            master=self.canvas_frame,
        )

        self.check_axes_colours()

        self.extvers = extvers
        self.coverage = {}
        widths = [1 / 3, 1] * 3
        self.fig_axes = self.fig.subplots(
            4,
            6,
            sharey=True,
            width_ratios=widths,
        )

        self.quality_frame = MultiQualityFrame(self.canvas_frame, extvers=self.extvers)
        self.quality_frame.grid(row=1, column=0, sticky="ew")

        self.set_aspect()
        self.plotted_images = dict()
        self.update_plots()

    def check_axes_colours(self):
        self.fig.set_facecolor("none")
        if ctk.get_appearance_mode() == "Dark":
            self.fig.canvas.get_tk_widget().config(bg=self.cget("bg_color")[-1])
            fg_colour = [
                a / 65535
                for a in self._root().winfo_rgb(
                    self._root().progress_status.cget("text_color")[-1]
                )
            ]
        if ctk.get_appearance_mode() == "Light":
            self.fig.canvas.get_tk_widget().config(bg=self.cget("bg_color")[0])
            fg_colour = [
                a / 65535
                for a in self._root().winfo_rgb(
                    self._root().progress_status.cget("text_color")[0]
                )
            ]

        mpl.rcParams["text.color"] = fg_colour
        mpl.rcParams["axes.labelcolor"] = fg_colour
        mpl.rcParams["xtick.color"] = fg_colour
        mpl.rcParams["ytick.color"] = fg_colour
        mpl.rcParams["axes.edgecolor"] = fg_colour

    def set_aspect(self, aspect_ratio=3):
        # a function which places a frame within a containing frame, and
        # then forces the inner frame to keep a specific aspect ratio

        def enforce_aspect_ratio(event):
            # when the pad window resizes, fit the content into it,
            # either by fixing the width or the height and then
            # adjusting the height or width based on the aspect ratio.

            other_heights = self.quality_frame.winfo_height()
            # start by using the width as the controlling dimension
            desired_width = event.width
            desired_height = int(event.width / aspect_ratio) + other_heights

            # if the window is too tall to fit, use the height as
            # the controlling dimension
            if desired_height > event.height:
                desired_height = event.height
                desired_width = int((event.height - other_heights) * aspect_ratio)

            # place the window, giving it an explicit size
            self.canvas_frame.place(
                in_=self.pad_frame,
                x=0,
                y=0,
                relwidth=desired_width / event.width,
                relheight=desired_height / event.height,
            )

        self.pad_frame.bind("<Configure>", enforce_aspect_ratio)

    def update_plots(self, extvers=None):
        import time

        t1 = time.perf_counter()
        self.check_axes_colours()

        if extvers != None:
            self.extvers = extvers
            self.quality_frame.reload_extvers(new_extvers=self.extvers)

        self.master.gal_id = self._root().current_gal_id.get()

        if self._root().plot_options["stretch"].lower() == "linear":
            self.stretch_fn = LinearStretch
        elif self._root().plot_options["stretch"].lower() == "square root":
            self.stretch_fn = SqrtStretch
        elif self._root().plot_options["stretch"].lower() == "logarithmic":
            self.stretch_fn = LogStretch

        for j, name in enumerate(["SCI", "CONTAM", "MODEL", "RESIDUALS"]):
            for i, ver in enumerate(self.extvers):
                if name + ver not in self.plotted_images.keys():
                    self.plotted_images[name + ver] = dict()
                self.plot_kernel(self.fig_axes[j, 2 * i], name, ver)
                self.plot_beam(self.fig_axes[j, (2 * i) + 1], name, ver)

        # print("T2:", time.perf_counter() - t1)

        self.fig.canvas.draw_idle()

        self.fig.canvas.get_tk_widget().grid(row=0, column=0, sticky="news")

        # print("T3:", time.perf_counter() - t1)

    def plot_kernel(self, ax, ext, extver):
        with pf.open(self.master.file_path) as hdul:
            try:
                data = hdul["KERNEL", extver].data

                if self._root().plot_options["limits"] == "grizli default":
                    vmax_kern = 1.1 * np.percentile(data, 99.5)
                    interval = ManualInterval(vmin=-0.1 * vmax_kern, vmax=vmax_kern)
                elif self._root().plot_options["limits"] == "Min-max":
                    interval = MinMaxInterval()
                else:
                    interval = PercentileInterval(
                        float(self._root().plot_options["limits"].replace("%", ""))
                    )

                norm = ImageNormalize(
                    data,
                    interval=interval,
                    stretch=self.stretch_fn(),
                )
                try:
                    self.plotted_images[ext + extver]["kernel"].set_data(data)
                    self.plotted_images[ext + extver]["kernel"].set_norm(norm)
                    self.plotted_images[ext + extver]["kernel"].set_cmap(
                        self._root().plot_options["cmap"]
                    )
                    self.plotted_images[ext + extver]["kernel"].set_visible(True)
                except Exception as e:
                    self.plotted_images[ext + extver]["kernel"] = ax.imshow(
                        data,
                        origin="lower",
                        cmap=self._root().plot_options["cmap"],
                        # aspect="auto"
                        norm=norm,
                        visible=True,
                        interpolation="nearest",
                    )
                ax.set_xticklabels("")
                ax.set_yticklabels("")
                ax.tick_params(axis="both", direction="in", top=True, right=True)
                if ax in self.fig_axes[:, 0]:
                    ax.set_ylabel(ext)
            except Exception as e:
                if "kernel" in self.plotted_images[ext + extver].keys():
                    self.plotted_images[ext + extver]["kernel"].set_visible(False)
                pass

    def plot_beam(self, ax, ext, extver):
        with pf.open(self.master.file_path) as hdul:
            try:
                if ext == "RESIDUALS":
                    data = hdul["SCI", extver].data
                    m = hdul["MODEL", extver].data
                else:
                    data = hdul[ext, extver].data
                    m = 0

                if ext == "SCI":
                    self.coverage[extver] = (
                        1
                        - np.sum(np.all((~np.isfinite(data)) | (data == 0), axis=0))
                        / data.shape[1]
                    )
                    self._root().current_gal_data[extver]["coverage"] = self.coverage[
                        extver
                    ]

                header = hdul["SCI", extver].header
                extent = [header["WMIN"], header["WMAX"], 0, data.shape[0]]

                if self._root().plot_options["limits"] == "grizli default":
                    wht_i = hdul["WHT", extver]
                    clip = wht_i.data > 0
                    if clip.sum() == 0:
                        clip = np.isfinite(wht_i.data)

                    avg_rms = 1 / np.median(np.sqrt(wht_i.data[clip]))
                    vmax = np.maximum(1.1 * np.percentile(data[clip], 98), 5 * avg_rms)
                    vmin = -0.1 * vmax
                    interval = ManualInterval(vmin=vmin, vmax=vmax)
                elif self._root().plot_options["limits"] == "Min-max":
                    interval = MinMaxInterval()
                else:
                    interval = PercentileInterval(
                        float(self._root().plot_options["limits"].replace("%", ""))
                    )

                norm = ImageNormalize(
                    data,
                    interval=interval,
                    stretch=self.stretch_fn(),
                )
                try:
                    self.plotted_images[ext + extver]["beam"].set_data(data - m)
                    self.plotted_images[ext + extver]["beam"].set_norm(norm)
                    self.plotted_images[ext + extver]["beam"].set_cmap(
                        self._root().plot_options["cmap"]
                    )
                    self.plotted_images[ext + extver]["beam"].set_visible(True)
                except:
                    self.plotted_images[ext + extver]["beam"] = ax.imshow(
                        data - m,
                        origin="lower",
                        cmap=self._root().plot_options["cmap"],
                        aspect="auto",
                        norm=norm,
                        extent=extent,
                        visible=True,
                        interpolation="nearest",
                    )
                ax.tick_params(axis="both", direction="in", top=True, right=True)

                if ax not in self.fig_axes[-1]:
                    ax.set_xticklabels("")
                    ax.set_yticklabels("")
                else:
                    ax.set_xlabel(r"$\lambda$ ($\mu$m) - " + extver.split(",")[0])
                try:
                    self.plotted_images[ext + extver]["beam_failed"].set_visible(False)
                except:
                    pass
            except KeyError as e:
                # print (e)
                try:
                    self.plotted_images[ext + extver]["beam_failed"].set_visible(True)
                except Exception as e:
                    # print (e)

                    self.plotted_images[ext + extver]["beam_failed"] = ax.text(
                        0.5,
                        0.5,
                        "No data",
                        transform=ax.transAxes,
                        ha="center",
                        va="center",
                        c=self._root().text_colour,
                    )
                try:
                    self.plotted_images[ext + extver]["beam"].set_visible(False)
                except:
                    pass
                self._root().current_gal_data[extver]["coverage"] = 0.0
                self.quality_frame.quality_menus[extver].set("Unusable")
                self.quality_frame.quality_menus[extver].configure(state="disabled")
            except Exception as e:
                print("beam:", e)
                pass


class MultiQualityFrame(ctk.CTkFrame):
    def __init__(self, master, extvers, **kwargs):
        super().__init__(master, **kwargs)
        self.columnconfigure((0, 1, 2, 3, 4, 5), weight=1)
        self.extvers = extvers
        self.quality_menus = {}

        self.possible_values = ["Excellent", "Good", "Poor", "Unusable"]

        for i, e in enumerate(
            self.extvers,
        ):
            label = ctk.CTkLabel(self, text="Beam Quality")
            label.grid(row=0, column=2 * i, padx=10, pady=(10, 0), sticky="e")
            q_menu = ctk.CTkOptionMenu(
                self,
                values=self.possible_values,
                command=self.save_current,
            )
            try:
                q_menu.set(self._root().current_gal_data[e]["quality"])
            except:
                q_menu.set("Good")
            q_menu.grid(row=0, column=2 * i + 1, padx=10, pady=(10, 0), sticky="w")
            self.quality_menus[e] = q_menu

        self.save_current()

    def keypress_select(self, event, key_maps):
        if event.char in key_maps:
            idx = (event.char == key_maps).nonzero()
            self._root().current_gal_data[self.extvers[int(idx[0])]][
                "quality"
            ] = self.possible_values[int(idx[1])]
            self.quality_menus[self.extvers[int(idx[0])]].set(
                self.possible_values[int(idx[1])]
            )

    def reload_extvers(self, new_extvers=None):
        if self.extvers != None:
            self.extvers = new_extvers

        for e, c in zip(self.extvers, self.quality_menus.values()):
            try:
                c.set(self._root().current_gal_data[e]["quality"])
            except:
                c.set("Good")
            c.configure(state="normal")

        self.save_current()

    def save_current(self, event=None):
        for v, q in zip(self.extvers, self.quality_menus.values()):
            if v not in self._root().current_gal_data.keys():
                self._root().current_gal_data[v] = {}
            self._root().current_gal_data[v]["quality"] = q.get()
